/*
 *  p3t.c
 *  
 *
 *  Created by Johannes Wilke on 2008/08/16.
 *  Copyright 2008 __MyCompanyName__. All rights reserved.
 *
 */

#include "p3t.h"

#include <stdio.h>
#include <list>
#include <map>
#include <string>
#include "gim.h"
#include <sys/stat.h>

#define OUTPUT_DEBUG_TEXT 1
#define OUTPUT_ERROR_TEXT 1

typedef unsigned long p3tOffsetType;
typedef unsigned long p3tSizeType;

struct p3tLocationEntry 
{
  p3tOffsetType offset;
  p3tSizeType size;
};

struct p3tHeader
{
  unsigned long magic;
  unsigned long version;
  struct p3tLocationEntry elements;
  struct p3tLocationEntry ids;
  struct p3tLocationEntry strings;
  struct p3tLocationEntry integers;
  struct p3tLocationEntry decimals;
  struct p3tLocationEntry files;
  unsigned long padding[2];
};

union p3tAttributeValue
{
  p3tLocationEntry location;
  p3tInt integer;
  p3tFloat decimal;
};

typedef unsigned long p3tAttributeType;

struct p3tAttributeHeader
{
  p3tOffsetType name_offset;
  p3tAttributeType type;
  p3tAttributeValue value;
};

typedef enum {
  P3T_GIM_FILE,
  P3T_JPG_FILE,
  P3T_VAG_FILE
} p3tFileType;

struct p3tAttribute
{
  p3tAttributeHeader header;
  unsigned char* file_data;
  p3tSizeType file_size;
  p3tFileType file_type;  
};

typedef std::list<p3tAttribute> p3tAttributeList;

struct p3tElementHeader
{
  p3tOffsetType name_offset;
  unsigned long no_attributes;
  p3tOffsetType parent_offset;
  p3tOffsetType previous_sibling_offset;
  p3tOffsetType next_sibling_offset;
  p3tOffsetType first_child_offset;
  p3tOffsetType last_child_offset;
};

struct p3tElement
{
  p3tElementHeader header;
  p3tAttributeList attributes;
  
  p3tOffsetType offset;
};
typedef std::map<p3tOffsetType, p3tElement> elementMap;

typedef std::map<p3tOffsetType, p3tString> stringMap;
typedef std::map<p3tOffsetType, p3tId> idMap;
typedef std::map<p3tOffsetType, p3tOffsetType> idParentOffsetMap;

// ================================ Ripping p3t FILE =========================

int loadP3THeader(FILE* p3t_file_handle, p3tHeader& p3theader)
{
  p3tSizeType no_header_bytes_to_read = sizeof(p3theader);
  p3tSizeType no_header_bytes_read = fread(&p3theader, 1, no_header_bytes_to_read, p3t_file_handle);
  if (no_header_bytes_read != no_header_bytes_to_read)
  {
    fclose(p3t_file_handle);
    return P3T_COULD_NOT_READ_HEADER;
  }
  
#ifdef OUTPUT_DEBUG_TEXT
  char magic_string[5];
  *((unsigned long*)magic_string) = p3theader.magic;
  magic_string[4] = 0;
  printf ("magic=%s\n", magic_string);
  printf ("version=%d\n", p3theader.version);
  printf ("element offset=%d size=%d\n", p3theader.elements.offset, p3theader.elements.size);
  printf ("id table offset=%d size=%d\n", p3theader.ids.offset, p3theader.ids.size);
  printf ("string table offset=%d size=%d\n", p3theader.strings.offset, p3theader.strings.size);
  printf ("integer array offset=%d size=%d\n", p3theader.integers.offset, p3theader.integers.size);
  printf ("float array offset=%d size=%d\n", p3theader.decimals.offset, p3theader.decimals.size);
  printf ("file table offset=%d size=%d\n", p3theader.files.offset, p3theader.files.size);
#endif
  
  return P3T_OK;
}

int loadP3TStringTable(FILE* p3t_file_handle, p3tHeader& p3theader, stringMap& strings)
{
#ifdef OUTPUT_DEBUG_TEXT
  printf("\n===== Reading Strings =====\n");
#endif
  
  char* string_table_buffer = new char[p3theader.strings.size];
  if (string_table_buffer == 0)
  {
    fclose(p3t_file_handle);
    return P3T_OUT_OF_MEMORY;
  }
  
  fseek(p3t_file_handle, p3theader.strings.offset, SEEK_SET);
  p3tSizeType no_string_table_bytes_read = fread(string_table_buffer, 1, p3theader.strings.size, p3t_file_handle);
  if (no_string_table_bytes_read != p3theader.strings.size)
  {
    delete[]string_table_buffer;
    fclose(p3t_file_handle);
    return P3T_COULD_NOT_READ_STRING_TABLE;
  }
  
  p3tOffsetType next_string_offset = 0;
  while (next_string_offset < p3theader.strings.size)
  {
    std::string next_string(&string_table_buffer[next_string_offset]);
    strings[next_string_offset] = next_string;
    
#ifdef OUTPUT_DEBUG_TEXT
    printf ("%d %d %s\n", next_string_offset, next_string.size(), next_string.c_str());
#endif
    
    next_string_offset += next_string.size() + 1;
  }
  delete[]string_table_buffer;
  
  return P3T_OK;
}

int loadP3TIdTable(FILE* p3t_file_handle, p3tHeader& p3theader, idMap& ids)
{
#ifdef OUTPUT_DEBUG_TEXT
  printf("\n===== Reading IDS =====\n");  
#endif
  
  char* id_table_buffer = new char[p3theader.ids.size];
  if (id_table_buffer == 0)
  {
    fclose(p3t_file_handle);
    return P3T_OUT_OF_MEMORY;
  }
  
  fseek(p3t_file_handle, p3theader.ids.offset, SEEK_SET);
  p3tSizeType no_id_table_bytes_read = fread(id_table_buffer, 1, p3theader.ids.size, p3t_file_handle);
  if (no_id_table_bytes_read != p3theader.ids.size)
  {
    delete[]id_table_buffer;
    fclose(p3t_file_handle);
    return P3T_COULD_NOT_READ_ID_TABLE;
  }
  
  p3tOffsetType next_id_offset = 0;
  while (next_id_offset < p3theader.ids.size)
  {
    p3tOffsetType offset = next_id_offset;
    
    // skip the elementID
    next_id_offset += 4;
    
    std::string id_text(&id_table_buffer[next_id_offset]);
    
    ids[offset] = id_text;
    
#ifdef OUTPUT_DEBUG_TEXT
    printf ("%d %d %s\n", offset, id_text.size(), id_text.c_str());
#endif
    
    next_id_offset += id_text.size() + 1;
  }
  delete[]id_table_buffer;
  
  return P3T_OK;
}

void extractRippedElements(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements);

int p3tRip(const char* p3t_filename, p3tTheme& p3t_theme)
{
  FILE* p3t_file_handle = fopen(p3t_filename, "rb");
  if (p3t_file_handle == 0)
  {
    return P3T_CANNOT_OPEN_FILE;
  }
  
  int result;
  
  struct p3tHeader p3theader;
  result = loadP3THeader(p3t_file_handle, p3theader);
  if (result != P3T_OK)
    return result;
  
  if (p3theader.magic != 0x50335446)
    return P3T_CORRUPT_P3T_FILE;
  
  stringMap strings;
  result = loadP3TStringTable(p3t_file_handle, p3theader, strings);
  if (result != P3T_OK)
    return result;
  
  idMap ids;
  result = loadP3TIdTable(p3t_file_handle, p3theader, ids);
  if (result != P3T_OK)
    return result;
  
  // ==================== Rip Elements with their attributes
  
  elementMap elements;
  
  fseek(p3t_file_handle, p3theader.elements.offset, SEEK_SET);
  
  p3tOffsetType next_element_offset = 0;
  while (next_element_offset < p3theader.elements.size)
  {
    p3tElement element;
    p3tOffsetType elementID = next_element_offset;
    
    fread(&element.header, 1, sizeof(element.header), p3t_file_handle);
    element.offset = elementID;
#ifdef OUTPUT_DEBUG_TEXT
    printf("\n===== Element %d =====\n name=%s no_atrr=%d prev_sib=%d next_sib=%d frst_cld=%d last_cld=%d ", elementID, 
           strings[element.header.name_offset].c_str(), 
           element.header.no_attributes,
           element.header.previous_sibling_offset, element.header.next_sibling_offset,
           element.header.first_child_offset, element.header.last_child_offset);
    if (element.header.parent_offset != (p3tOffsetType)-1)
    {
      printf("parent=%s\n", strings[elements[element.header.parent_offset].header.name_offset].c_str());
    }
    else
      printf("ROOT\n");
#endif
    
    next_element_offset += sizeof(element.header);
    
    unsigned int attribute_index;
    for (attribute_index = 0; attribute_index < element.header.no_attributes; ++attribute_index)
    {
      struct p3tAttribute attribute;
      fread(&attribute.header, 1, sizeof(attribute.header), p3t_file_handle);
      
#ifdef OUTPUT_DEBUG_TEXT
      printf("%d AttributeName=%s type=%d value=", next_element_offset, strings[attribute.header.name_offset].c_str(), attribute.header.type);
#endif
      next_element_offset += sizeof(attribute.header);
      
      switch (attribute.header.type)
      {
        case 1: 
#ifdef OUTPUT_DEBUG_TEXT
          printf("%d\n", attribute.header.value.integer);
#endif
          break;
          
        case 2: 
#ifdef OUTPUT_DEBUG_TEXT
          printf("%f\n", attribute.header.value.decimal); 
#endif
          
          break;
          
        case 3: 
#ifdef OUTPUT_DEBUG_TEXT
          printf("%s [%d]\n", strings[attribute.header.value.location.offset].c_str(), attribute.header.value.location.size); 
#endif
          
          break;
          
        case 6: 
#ifdef OUTPUT_DEBUG_TEXT
          printf("%d %d\n", attribute.header.value.location.offset, attribute.header.value.location.size);
#endif
          
          p3tOffsetType current_file_position = ftell(p3t_file_handle);
          attribute.file_data = new unsigned char[attribute.header.value.location.size];
          fseek(p3t_file_handle, p3theader.files.offset + attribute.header.value.location.offset, SEEK_SET);
          fread(attribute.file_data, 1, attribute.header.value.location.size, p3t_file_handle);
          fseek(p3t_file_handle, current_file_position, SEEK_SET);
          
          if (strings[element.header.name_offset] == "se")
          {
            //static std::string vag_filename;
            //vag_filename = (vag_filename+"1");
            
            attribute.file_type = P3T_VAG_FILE;
            attribute.file_size = attribute.header.value.location.size;
            
#ifdef OUTPUT_DEBUG_TEXT
            printf("    VAG sound\n");
#endif
            
            //FILE* vag_file_handle = fopen((vag_filename+".vag").c_str(), "wb");
            //fwrite(attribute.file_data, 1, attribute.header.value.location.size, vag_file_handle);
            //fclose(vag_file_handle);
          }
          else if (strings[element.header.name_offset] == "bgimage")
          {
            //static std::string jpg_filename;
            //jpg_filename = (jpg_filename+"0");
            
            attribute.file_type = P3T_JPG_FILE;
            attribute.file_size = attribute.header.value.location.size;
            
#ifdef OUTPUT_DEBUG_TEXT
            printf("    JPG image\n");
#endif
            
            //FILE* jpg_file_handle = fopen((jpg_filename+".jpg").c_str(), "wb");
            //fwrite(attribute.file_data, 1, attribute.header.value.location.size, jpg_file_handle);
            //fclose(jpg_file_handle);
          }
          else
          {
            attribute.file_type = P3T_GIM_FILE;
            
            void* uncompressed_buffer;
            int uncompressed_size = uncompressGIMBuffer(attribute.file_data, attribute.header.value.location.size, &uncompressed_buffer);
            
            delete[](char*)attribute.file_data;
            attribute.file_data = 0;
            
            if (uncompressed_size < 0)
            {
              attribute.file_size = 0;
              
#ifdef OUTPUT_ERROR_TEXT
              printf("*** Element=%s AttributeName=%s Type=%d\n", 
                     strings[element.header.name_offset].c_str(), 
                     strings[attribute.header.name_offset].c_str(), attribute.header.type); 
#ifndef OUTPUT_DEBUG_TEXT
              printf("\n");
#else
              printf("*** COULD NOT UNCOMPRESS\n");
#endif
              
#endif
            }
            else
            {              
              attribute.file_data = new unsigned char[uncompressed_size];
              
              if (attribute.file_data == 0)
              {
                attribute.file_size = 0;
                
#ifdef OUTPUT_ERROR_TEXT
                printf("*** COULD NOT ALLOCATE MEMORY FOR UNCOMPRESSED FILE\n");
#endif
              }
              else
              {
                memcpy(attribute.file_data, uncompressed_buffer, uncompressed_size);
                attribute.file_size = uncompressed_size;
              }
              
#ifdef OUTPUT_DEBUG_TEXT
              printf("    uncommpressedsize=%d\n", uncompressed_size);
#endif
              
              delete[](char*)uncompressed_buffer;
            }
          }
          
          break;
          
          case 7: 
#ifdef OUTPUT_DEBUG_TEXT
          printf("%s[%d]\n", ids[attribute.header.value.location.offset].c_str(), attribute.header.value.location.size); 
#endif
          
          break;
          
          default: 
#ifdef OUTPUT_DEBUG_TEXT
          printf("N/A\n");
#endif
          
          ;
      }      
      element.attributes.push_back(attribute);
    }
    
    elements[elementID] = element;
  }
  
  fclose(p3t_file_handle);
  
  extractRippedElements(p3t_theme, strings, ids, elements);
  
  return P3T_OK;
}

void extractInfoElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element);
void extractLocalizedInfoElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element);
void extractSoundEffectElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element);
void extractIconElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element);
void extractPointerElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element);
void extractNotificationElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element);
void extractBGImageElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element);
void extractFontElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element);
void extractColorElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element);

#define GET_ATTRIBUTE_INT_VALUE attribute.header.value.integer
#define GET_ATTRIBUTE_STRING_VALUE strings[attribute.header.value.location.offset]
#define GET_ATTRIBUTE_FILE_DATA attribute.file_data
#define GET_ATTRIBUTE_FILE_SIZE attribute.file_size
#define GET_ATTRIBUTE_ID_VALUE ids[attribute.header.value.location.offset]

void extractRippedElements(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements)
{  
  p3t_theme.bgimagetable.bgimage.clear();
  p3t_theme.icontable.icon.clear();
  p3t_theme.pointertable.pointer.clear();
  
  // Extract Elements and populate p3tTheme structure
  for (elementMap::const_iterator element_iterator = elements.begin(); element_iterator != elements.end(); ++element_iterator)
  {
    const p3tElement& element = (const p3tElement&)element_iterator->second;
    
    if (strings[element.header.name_offset] == "info")
    {
      extractInfoElement(p3t_theme, strings, ids, elements, element);
    }
    else if (strings[element.header.name_offset] == "localizedinfo")
    {
      extractLocalizedInfoElement(p3t_theme, strings, ids, elements, element);
    }
    else if (strings[element.header.name_offset] == "se")
    {
      extractSoundEffectElement(p3t_theme, strings, ids, elements, element);
    }
    else if (strings[element.header.name_offset] == "icon")
    {
      extractIconElement(p3t_theme, strings, ids, elements, element);
    }
    else if (strings[element.header.name_offset] == "pointer")
    {
      extractPointerElement(p3t_theme, strings, ids, elements, element);
    }
    else if (strings[element.header.name_offset] == "notification")
    {
      extractNotificationElement(p3t_theme, strings, ids, elements, element);
    }
    else if (strings[element.header.name_offset] == "bgimage")
    {
      extractBGImageElement(p3t_theme, strings, ids, elements, element);
    }
    else if (strings[element.header.name_offset] == "font")
    {
      extractFontElement(p3t_theme, strings, ids, elements, element);
    }
    else if (strings[element.header.name_offset] == "color")
    {
      extractColorElement(p3t_theme, strings, ids, elements, element);
    }
  }
}

void extractInfoElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element)
{
  for (p3tAttributeList::const_iterator attribute_iterator = element.attributes.begin(); attribute_iterator != element.attributes.end(); ++attribute_iterator)
  {
    const p3tAttribute& attribute = (const p3tAttribute&)*attribute_iterator; 
    std::string attribute_name = strings[attribute.header.name_offset];
    
    if (attribute_name == "comment")
    {
      p3t_theme.infotable.info.comment = GET_ATTRIBUTE_STRING_VALUE;
    }
    else if (attribute_name == "previewsize")
    {
      p3t_theme.infotable.info.previewsize = GET_ATTRIBUTE_INT_VALUE;
    }
    else if (attribute_name == "name")
    {
      p3t_theme.infotable.info.name = GET_ATTRIBUTE_STRING_VALUE;
    }
    else if (attribute_name == "author")
    {
      p3t_theme.infotable.info.author = GET_ATTRIBUTE_STRING_VALUE;
    }
    else if (attribute_name == "url")
    {
      p3t_theme.infotable.info.url = GET_ATTRIBUTE_STRING_VALUE;
    }
    else if (attribute_name == "authoriconsize")
    {
      p3t_theme.infotable.info.authoriconsize = GET_ATTRIBUTE_INT_VALUE;
    }
    else if (attribute_name == "authoricon")
    {
      p3t_theme.infotable.info.authoricon.file_data = GET_ATTRIBUTE_FILE_DATA;
      p3t_theme.infotable.info.authoricon.file_size = GET_ATTRIBUTE_FILE_SIZE;
    }
    else if (attribute_name == "iconsize")
    {
      p3t_theme.infotable.info.iconsize = GET_ATTRIBUTE_INT_VALUE;
    }
    else if (attribute_name == "version")
    {
      p3t_theme.infotable.info.version = GET_ATTRIBUTE_STRING_VALUE;
    }
    else if (attribute_name == "mtime")
    {
      p3t_theme.infotable.info.mtime = GET_ATTRIBUTE_STRING_VALUE;
    }
    else if (attribute_name == "genre")
    {
      p3t_theme.infotable.info.genre = GET_ATTRIBUTE_STRING_VALUE;
    }
    else if (attribute_name == "preview")
    {
      p3t_theme.infotable.info.preview.file_data = GET_ATTRIBUTE_FILE_DATA;
      p3t_theme.infotable.info.preview.file_size = GET_ATTRIBUTE_FILE_SIZE;
    }
    else if (attribute_name == "icon")
    {
      p3t_theme.infotable.info.icon.file_data = GET_ATTRIBUTE_FILE_DATA;
      p3t_theme.infotable.info.icon.file_size = GET_ATTRIBUTE_FILE_SIZE;
    }
    else
    {
#ifdef OUTPUT_ERROR_TEXT
      printf("*** Unknown InfoElement attribute: %s ***\n", attribute_name.c_str());
#endif
    }
  }
}

void extractLocalizedInfoElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element)
{
  p3tLocalizedInfoElement local_info;
  
  for (p3tAttributeList::const_iterator attribute_iterator = element.attributes.begin(); attribute_iterator != element.attributes.end(); ++attribute_iterator)
  {
    const p3tAttribute& attribute = (const p3tAttribute&)*attribute_iterator; 
    std::string attribute_name = strings[attribute.header.name_offset];
    
    if (attribute_name == "locale")
    {
      local_info.locale = GET_ATTRIBUTE_ID_VALUE;
    }
    else if (attribute_name == "comment")
    {
      local_info.comment = GET_ATTRIBUTE_STRING_VALUE;
    }
    else if (attribute_name == "name")
    {
      local_info.name = GET_ATTRIBUTE_STRING_VALUE;
    }
    else if (attribute_name == "author")
    {
      local_info.author = GET_ATTRIBUTE_STRING_VALUE;
    }
    else if (attribute_name == "url")
    {
      local_info.url = GET_ATTRIBUTE_STRING_VALUE;
    }
    else if (attribute_name == "version")
    {
      local_info.version = GET_ATTRIBUTE_STRING_VALUE;
    }
    else if (attribute_name == "genre")
    {
      local_info.genre = GET_ATTRIBUTE_STRING_VALUE;
    }
    else
    {
#ifdef OUTPUT_ERROR_TEXT
      printf("*** Unknown LocalizedInfoElement attribute: %s\n", attribute_name.c_str());
#endif
    }
  }
  
  p3t_theme.infotable.localizedinfo.push_back(local_info);
}

void extractSoundEffectElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element)
{
  p3tSoundEffectElement se;
  se.left.file_data = 0;
  se.left.file_size = 0;
  se.right.file_data = 0;
  se.right.file_size = 0;
  
  for (p3tAttributeList::const_iterator attribute_iterator = element.attributes.begin(); attribute_iterator != element.attributes.end(); ++attribute_iterator)
  {
    const p3tAttribute& attribute = (const p3tAttribute&)*attribute_iterator; 
    std::string attribute_name = strings[attribute.header.name_offset];
    
    if (attribute_name == "src")
    {
      se.left.file_data = GET_ATTRIBUTE_FILE_DATA;
      se.left.file_size = GET_ATTRIBUTE_FILE_SIZE;
      se.right.file_data = 0;
      se.right.file_size = 0;
    }
    else if (attribute_name == "left")
    {
      se.left.file_data = GET_ATTRIBUTE_FILE_DATA;
      se.left.file_size = GET_ATTRIBUTE_FILE_SIZE;
    }
    else if (attribute_name == "right")
    {
      se.right.file_data = GET_ATTRIBUTE_FILE_DATA;
      se.right.file_size = GET_ATTRIBUTE_FILE_SIZE;
    }
    else if (attribute_name == "id")
    {
      se.id = GET_ATTRIBUTE_ID_VALUE;
    }
    else
    {
#ifdef OUTPUT_ERROR_TEXT
      printf("*** Unknown SEElement attribute: %s\n", attribute_name.c_str());
#endif
    }
  }
  
  p3t_theme.setable.se.push_back(se);
}

void extractIconElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element)
{
  p3tIconElement icon;
  icon.src.file_data = 0;
  icon.src.file_size = 0;
  icon.size = 0;
  
  for (p3tAttributeList::const_iterator attribute_iterator = element.attributes.begin(); attribute_iterator != element.attributes.end(); ++attribute_iterator)
  {
    const p3tAttribute& attribute = (const p3tAttribute&)*attribute_iterator; 
    std::string attribute_name = strings[attribute.header.name_offset];
    
    if (attribute_name == "src")
    {
      icon.src.file_data = GET_ATTRIBUTE_FILE_DATA;
      icon.src.file_size = GET_ATTRIBUTE_FILE_SIZE;
    }
    else if (attribute_name == "id")
    {
      icon.id = GET_ATTRIBUTE_ID_VALUE;
    }
    else if (attribute_name == "size")
    {
      icon.size = GET_ATTRIBUTE_INT_VALUE;
    }
    else
    {
#ifdef OUTPUT_ERROR_TEXT
      printf("*** Unknown IconElement attribute: %s\n", attribute_name.c_str());
#endif
    }
  }
  
  p3t_theme.icontable.icon.push_back(icon);
}

void extractPointerElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element)
{
  p3tPointerElement pointer;
  pointer.src.file_data = 0;
  pointer.src.file_size = 0;
  pointer.size = 0;
  
  for (p3tAttributeList::const_iterator attribute_iterator = element.attributes.begin(); attribute_iterator != element.attributes.end(); ++attribute_iterator)
  {
    const p3tAttribute& attribute = (const p3tAttribute&)*attribute_iterator; 
    std::string attribute_name = strings[attribute.header.name_offset];
    
    if (attribute_name == "base_x")
    {
      pointer.base_x = GET_ATTRIBUTE_INT_VALUE;
    }
    else  if (attribute_name == "src")
    {
      pointer.src.file_data = GET_ATTRIBUTE_FILE_DATA;
      pointer.src.file_size = GET_ATTRIBUTE_FILE_SIZE;
    }
    else if (attribute_name == "size")
    {
      pointer.size = GET_ATTRIBUTE_INT_VALUE;
    }
    else if (attribute_name == "id")
    {
      pointer.id = GET_ATTRIBUTE_ID_VALUE;
    }
    else if (attribute_name == "base_y")
    {
      pointer.base_y = GET_ATTRIBUTE_INT_VALUE;
    }
    else
    {
#ifdef OUTPUT_ERROR_TEXT
      printf("*** Unknown PointerElement attribute: %s\n", attribute_name.c_str());
#endif
    }
  }
  
  p3t_theme.pointertable.pointer.push_back(pointer);
}

void extractNotificationElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element)
{
  for (p3tAttributeList::const_iterator attribute_iterator = element.attributes.begin(); attribute_iterator != element.attributes.end(); ++attribute_iterator)
  {
    const p3tAttribute& attribute = (const p3tAttribute&)*attribute_iterator; 
    std::string attribute_name = strings[attribute.header.name_offset];
    
    if (attribute_name == "src")
    {
      p3t_theme.notification.src.file_data = GET_ATTRIBUTE_FILE_DATA;
      p3t_theme.notification.src.file_size = GET_ATTRIBUTE_FILE_SIZE;
    }
    else if (attribute_name == "size")
    {
      p3t_theme.notification.size = GET_ATTRIBUTE_INT_VALUE;
    }
    else
    {
#ifdef OUTPUT_ERROR_TEXT
      printf("*** Unknown NotificationElement attribute: %s\n", attribute_name.c_str());
#endif
    }    
  }
}

void extractBGImageElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element)
{
  p3tBGImageElement bgimage;
  
  for (p3tAttributeList::const_iterator attribute_iterator = element.attributes.begin(); attribute_iterator != element.attributes.end(); ++attribute_iterator)
  {
    const p3tAttribute& attribute = (const p3tAttribute&)*attribute_iterator; 
    std::string attribute_name = strings[attribute.header.name_offset];
    
    if (attribute_name == "hd")
    {
      bgimage.hd.file_data = GET_ATTRIBUTE_FILE_DATA;
      bgimage.hd.file_size = GET_ATTRIBUTE_FILE_SIZE;
    }
    else if (attribute_name == "sd")
    {
      bgimage.sd.file_data = GET_ATTRIBUTE_FILE_DATA;
      bgimage.sd.file_size = GET_ATTRIBUTE_FILE_SIZE;
    }
    else
    {
#ifdef OUTPUT_ERROR_TEXT
      printf("*** Unknown BGImageElement attribute: %s\n", attribute_name.c_str());
#endif
    }
  }
  
  p3t_theme.bgimagetable.bgimage.push_back(bgimage);
}

void extractFontElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element)
{
  for (p3tAttributeList::const_iterator attribute_iterator = element.attributes.begin(); attribute_iterator != element.attributes.end(); ++attribute_iterator)
  {
    const p3tAttribute& attribute = (const p3tAttribute&)*attribute_iterator; 
    std::string attribute_name = strings[attribute.header.name_offset];
    
    if (attribute_name == "selection")
    {
      p3t_theme.font.selection = GET_ATTRIBUTE_INT_VALUE;
    }
    else
    {
#ifdef OUTPUT_ERROR_TEXT
      printf("*** Unknown FontElement attribute: %s\n", attribute_name.c_str());
#endif
    }
  }
}

void extractColorElement(p3tTheme& p3t_theme, stringMap& strings, idMap& ids, elementMap& elements, const p3tElement& element)
{
  for (p3tAttributeList::const_iterator attribute_iterator = element.attributes.begin(); attribute_iterator != element.attributes.end(); ++attribute_iterator)
  {
    const p3tAttribute& attribute = (const p3tAttribute&)*attribute_iterator; 
    std::string attribute_name = strings[attribute.header.name_offset];
    
    if (attribute_name == "selection")
    {
      p3t_theme.color.selection = GET_ATTRIBUTE_INT_VALUE;
    }
    else
    {
#ifdef OUTPUT_ERROR_TEXT
      printf("*** Unknown ColorElement attribute: %s\n", attribute_name.c_str());
#endif
    }
  }
}

// ================================ Building p3t FILE =========================

#define NO_OFFSET_REFERENCE ((p3tOffsetType)-1)

stringMap strings;
p3tOffsetType string_offset = 0;
idMap ids;
idParentOffsetMap idParents;
p3tOffsetType ids_offset = 0;
p3tOffsetType file_offset = 0;
elementMap elements;
p3tOffsetType element_offset = 0;

p3tOffsetType buildString(std::string string)
{
  for (stringMap::const_iterator string_iterator = strings.begin(); string_iterator != strings.end(); ++string_iterator)
  {
    if (string_iterator->second == string)
      return string_iterator->first;
  }
  
  p3tOffsetType new_string_offset = string_offset;
  strings[string_offset] = string;
  string_offset += string.size() + 1;
  
  return new_string_offset;
}

p3tElement& buildElement(std::string name, p3tOffsetType parent_offset, p3tOffsetType previous_sibling_offset)
{
  p3tElement element;
  
  element.header.name_offset = buildString(name);
  element.header.no_attributes = 0;
  element.header.parent_offset = parent_offset;
  element.header.previous_sibling_offset = previous_sibling_offset;
  element.header.next_sibling_offset = NO_OFFSET_REFERENCE;
  element.header.first_child_offset = NO_OFFSET_REFERENCE;
  element.header.last_child_offset = NO_OFFSET_REFERENCE;
  
  element.offset = element_offset;
  
  elements[element_offset] = element;
  elementMap::iterator element_iterator = elements.find(element_offset);
  
  element_offset += sizeof(element.header);
  
  return element_iterator->second;
}

void buildIntegerAttribute(p3tElement& element, std::string name, const p3tInt value)
{
  p3tAttribute attribute;
  
  attribute.header.name_offset = buildString(name);
  attribute.header.type = 1;
  attribute.header.value.integer = value;
  
  element.attributes.push_back(attribute);
  
  element_offset += sizeof(p3tAttributeHeader);
}

void buildFloatAttribute(p3tElement& element, std::string name, const p3tFloat value)
{
  p3tAttribute attribute;
  
  attribute.header.name_offset = buildString(name);
  attribute.header.type = 2;
  attribute.header.value.decimal = value;
  
  element.attributes.push_back(attribute);
  element_offset += sizeof(p3tAttributeHeader);
}

void buildStringAttribute(p3tElement& element, std::string name, const p3tString& value)
{
  if (value.size() == 0)
    return;
  
  p3tAttribute attribute;
  
  
  attribute.header.name_offset = buildString(name);
  attribute.header.type = 3;
  attribute.header.value.location.offset = buildString(value);
  attribute.header.value.location.size = value.size();
  
  element.attributes.push_back(attribute);
  element_offset += sizeof(p3tAttributeHeader);
}

void buildFileAttribute(p3tElement& element, std::string name, unsigned char* file_data, p3tSizeType file_size)
{
  p3tAttribute attribute;
  
  attribute.header.name_offset = buildString(name);
  
  unsigned char* compressed_data;
  p3tSizeType compressed_data_size;
  
  if (strings[element.header.name_offset] == "bgimage")
  {
    attribute.file_type = P3T_JPG_FILE;
    
    compressed_data = file_data;
    compressed_data_size = file_size;
  }
  else if (strings[element.header.name_offset] == "se")
  {
    attribute.file_type = P3T_VAG_FILE;
    
    compressed_data = file_data;
    compressed_data_size = file_size;
  }
  else
  {
    attribute.file_type = P3T_GIM_FILE;
    
    compressed_data = new unsigned char[file_size];
    if (compressed_data == 0)
    {
#ifdef ERROR_OUTPUT_TEXT
      printf("*** OUT OF MEMORY: allocate compressed buffer in buildFileAttribute\n");
#endif
      return;
    }
    
    compressed_data_size = compressGIMBuffer(file_data, file_size, compressed_data, file_size);
    if (compressed_data_size < 0)
    {
      delete[]compressed_data;
#ifdef ERROR_OUTPUT_TEXT
      printf("*** COULD NOT COMPRESS GIM FILE in buildFileAttribute\n");
#endif
      return;    
    }    
  }
  
  attribute.header.type = 6;
  attribute.header.value.location.offset = file_offset;
  attribute.header.value.location.size = compressed_data_size;
  attribute.file_data = compressed_data;
  attribute.file_size = compressed_data_size;
  file_offset += compressed_data_size;
  file_offset = ((file_offset + 15) / 16) * 16;
  
  element.attributes.push_back(attribute);
  element_offset += sizeof(p3tAttributeHeader);
}

void buildIdAttribute(p3tElement& element, std::string name, const p3tId& value)
{
  if (value.size() == 0)
    return;
  
  p3tAttribute attribute;
  
  attribute.header.name_offset = buildString(name);
  attribute.header.type = 7;
  ids[ids_offset] = value;
  idParents[ids_offset] = element.offset;
  attribute.header.value.location.offset = ids_offset;
  attribute.header.value.location.size = 0;
  ids_offset += value.size() + 1 + 4;
  
  element.attributes.push_back(attribute);
  element_offset += sizeof(p3tAttributeHeader);
}

int p3tBuild(const char* p3t_filename, p3tTheme& p3t_theme)
{  
  strings.clear();
  string_offset = 0;
  ids.clear();
  ids_offset = 0;
  file_offset = 0;
  elements.clear();
  element_offset = 0;
  
  p3tElement& theme = (p3tElement&)buildElement(std::string("theme"), NO_OFFSET_REFERENCE, NO_OFFSET_REFERENCE);
  p3tElement& infotable = (p3tElement&)buildElement(std::string("infotable"), theme.offset, NO_OFFSET_REFERENCE);
  theme.header.first_child_offset = infotable.offset;
  
  p3tElement& info = (p3tElement&)buildElement(std::string("info"), infotable.offset, NO_OFFSET_REFERENCE);
  infotable.header.first_child_offset = info.offset;
  if (p3t_theme.infotable.info.comment != "")
    buildStringAttribute(info, std::string("comment"), p3t_theme.infotable.info.comment);
  if (p3t_theme.infotable.info.preview.file_data)
    buildIntegerAttribute(info, std::string("previewsize"), p3t_theme.infotable.info.previewsize);
  if (p3t_theme.infotable.info.name != "")
    buildStringAttribute(info, std::string("name"), p3t_theme.infotable.info.name);
  if (p3t_theme.infotable.info.author != "")
    buildStringAttribute(info, std::string("author"), p3t_theme.infotable.info.author);
  if (p3t_theme.infotable.info.url != "")
    buildStringAttribute(info, std::string("url"), p3t_theme.infotable.info.url);
  if (p3t_theme.infotable.info.authoricon.file_data)
  {
    buildIntegerAttribute(info, std::string("authoriconsize"), p3t_theme.infotable.info.authoriconsize);
    buildFileAttribute(info, std::string("authoricon"), p3t_theme.infotable.info.authoricon.file_data, p3t_theme.infotable.info.authoricon.file_size);    
  }
  if (p3t_theme.infotable.info.icon.file_data)
    buildIntegerAttribute(info, std::string("iconsize"), p3t_theme.infotable.info.iconsize);
  if (p3t_theme.infotable.info.version != "")
    buildStringAttribute(info, std::string("version"), p3t_theme.infotable.info.version);
  if (p3t_theme.infotable.info.mtime != "")
    buildStringAttribute(info, std::string("mtime"), p3t_theme.infotable.info.mtime);
  if (p3t_theme.infotable.info.genre != "")
    buildStringAttribute(info, std::string("genre"), p3t_theme.infotable.info.genre);
  if (p3t_theme.infotable.info.preview.file_data)
    buildFileAttribute(info, std::string("preview"), p3t_theme.infotable.info.preview.file_data, p3t_theme.infotable.info.preview.file_size);
  if (p3t_theme.infotable.info.icon.file_data)
    buildFileAttribute(info, std::string("icon"), p3t_theme.infotable.info.icon.file_data, p3t_theme.infotable.info.icon.file_size);
  
  p3tElement* previous_element = 0;
  p3tOffsetType sibling_offset = info.offset;
  for (p3tLocalizedInfoList::const_iterator localizedinfo_iterator = p3t_theme.infotable.localizedinfo.begin(); localizedinfo_iterator != p3t_theme.infotable.localizedinfo.end(); ++localizedinfo_iterator)
  {
    if ((localizedinfo_iterator->comment == "") && (localizedinfo_iterator->name == "") &&
        (localizedinfo_iterator->author == "") && (localizedinfo_iterator->url == "") &&
        (localizedinfo_iterator->version == "") && (localizedinfo_iterator->genre == ""))
      continue;
    
    p3tElement& localizedinfo = (p3tElement&)buildElement(std::string("localizedinfo"), infotable.offset, sibling_offset);
    infotable.header.last_child_offset = localizedinfo.offset;
    if (previous_element)
    {
      previous_element->header.next_sibling_offset = localizedinfo.offset;
    }
    else
      info.header.next_sibling_offset = localizedinfo.offset;
    
    buildIdAttribute(localizedinfo, std::string("locale"), localizedinfo_iterator->locale);
    buildStringAttribute(localizedinfo, std::string("comment"), localizedinfo_iterator->comment);
    buildStringAttribute(localizedinfo, std::string("name"), localizedinfo_iterator->name);    
    buildStringAttribute(localizedinfo, std::string("author"), localizedinfo_iterator->author);    
    buildStringAttribute(localizedinfo, std::string("url"), localizedinfo_iterator->url);    
    buildStringAttribute(localizedinfo, std::string("version"), localizedinfo_iterator->version);    
    buildStringAttribute(localizedinfo, std::string("genre"), localizedinfo_iterator->genre);    
    
    sibling_offset = localizedinfo.offset;
    
    previous_element = &localizedinfo;
  }
  
  p3tElement& setable = (p3tElement&)buildElement(std::string("setable"), theme.offset, infotable.offset);
  infotable.header.next_sibling_offset = setable.offset;
  sibling_offset = NO_OFFSET_REFERENCE;
  previous_element = 0;
  for (p3tSoundEffectList::const_iterator se_iterator = p3t_theme.setable.se.begin(); se_iterator != p3t_theme.setable.se.end(); ++se_iterator)
  {
    if ((se_iterator->left.file_data) && (se_iterator->right.file_data))
    {
      p3tElement& se = (p3tElement&)buildElement(std::string("se"), setable.offset, sibling_offset);
      if (previous_element)
      {
        previous_element->header.next_sibling_offset = se.offset;
      }
      else
        setable.header.first_child_offset = se.offset;
      setable.header.last_child_offset = se.offset;
      
      buildFileAttribute(se, std::string("left"), se_iterator->left.file_data, se_iterator->left.file_size);
      buildIdAttribute(se, std::string("id"), se_iterator->id);
      buildFileAttribute(se, std::string("right"), se_iterator->right.file_data, se_iterator->right.file_size);
      sibling_offset = se.offset;
      
      previous_element = &se;
    }
    else if (se_iterator->left.file_data)
    {
      p3tElement& se = (p3tElement&)buildElement(std::string("se"), setable.offset, sibling_offset);
      if (previous_element)
      {
        previous_element->header.next_sibling_offset = se.offset;
      }
      else
        setable.header.first_child_offset = se.offset;
      setable.header.last_child_offset = se.offset;
      
      buildFileAttribute(se, std::string("src"), se_iterator->left.file_data, se_iterator->left.file_size);
      buildIdAttribute(se, std::string("id"), se_iterator->id);
      sibling_offset = se.offset;
      
      previous_element = &se;
    }
    else if (se_iterator->right.file_data)
    {
      p3tElement& se = (p3tElement&)buildElement(std::string("se"), setable.offset, sibling_offset);
      if (previous_element)
      {
        previous_element->header.next_sibling_offset = se.offset;
      }
      else
        setable.header.first_child_offset = se.offset;
      setable.header.last_child_offset = se.offset;
      
      buildFileAttribute(se, std::string("src"), se_iterator->right.file_data, se_iterator->right.file_size);
      buildIdAttribute(se, std::string("id"), se_iterator->id);
      sibling_offset = se.offset;
      
      previous_element = &se;
    }
  }
  
  p3tElement& icontable = (p3tElement&)buildElement(std::string("icontable"), theme.offset, setable.offset);
  setable.header.next_sibling_offset = icontable.offset;
  sibling_offset = NO_OFFSET_REFERENCE;
  previous_element = 0;
  for (p3tIconList::const_iterator icon_iterator = p3t_theme.icontable.icon.begin(); icon_iterator != p3t_theme.icontable.icon.end(); ++icon_iterator)
  {
    if (icon_iterator->src.file_data)
    {
      p3tElement& icon = (p3tElement&)buildElement(std::string("icon"), icontable.offset, sibling_offset);
      if (previous_element)
      {
        previous_element->header.next_sibling_offset = icon.offset;
      }
      else
        icontable.header.first_child_offset = icon.offset;
      icontable.header.last_child_offset = icon.offset;
      
      buildFileAttribute(icon, std::string("src"), icon_iterator->src.file_data, icon_iterator->src.file_size);
      buildIdAttribute(icon, std::string("id"), icon_iterator->id);
      buildIntegerAttribute(icon, std::string("size"), icon_iterator->size);    
      sibling_offset = icon.offset;
      
      previous_element = &icon;
    }
  }
  
  p3tElement& pointertable = (p3tElement&)buildElement(std::string("pointertable"), theme.offset, icontable.offset);
  icontable.header.next_sibling_offset = pointertable.offset;
  sibling_offset = NO_OFFSET_REFERENCE;
  previous_element = 0;
  for (p3tPointerList::const_iterator pointer_iterator = p3t_theme.pointertable.pointer.begin(); pointer_iterator != p3t_theme.pointertable.pointer.end(); ++pointer_iterator)
  {
    if (pointer_iterator->src.file_data)
    {
      p3tElement& pointer = (p3tElement&)buildElement(std::string("pointer"), pointertable.offset, sibling_offset);
      if (previous_element)
      {
        previous_element->header.next_sibling_offset = pointer.offset;
      }
      else
        pointertable.header.first_child_offset = pointer.offset;
      pointertable.header.last_child_offset = pointer.offset;
      
      buildIntegerAttribute(pointer, std::string("base_x"), pointer_iterator->base_x);
      buildFileAttribute(pointer, std::string("src"), pointer_iterator->src.file_data, pointer_iterator->src.file_size);
      buildIntegerAttribute(pointer, std::string("size"), pointer_iterator->size);    
      buildIdAttribute(pointer, std::string("id"), pointer_iterator->id);
      buildIntegerAttribute(pointer, std::string("base_y"), pointer_iterator->base_y);
      sibling_offset = pointer.offset;
      
      previous_element = &pointer;
    }
  }
  
  p3tElement* pointer_notification = &pointertable;
  if (p3t_theme.notification.src.file_data)
  {
    p3tElement& notification = (p3tElement&)buildElement(std::string("notification"), theme.offset, pointertable.offset);
    pointer_notification = &notification;
    pointertable.header.next_sibling_offset = notification.offset;
    buildFileAttribute(notification, std::string("src"), p3t_theme.notification.src.file_data, p3t_theme.notification.src.file_size);
    buildIntegerAttribute(notification, std::string("size"), p3t_theme.notification.size);    
  }
  
  p3tElement& bgimagetable = (p3tElement&)buildElement(std::string("bgimagetable"), theme.offset, pointer_notification->offset);
  pointer_notification->header.next_sibling_offset = bgimagetable.offset;
  sibling_offset = NO_OFFSET_REFERENCE;
  previous_element = 0;
  for (p3tBGImageList::const_iterator bgimage_iterator = p3t_theme.bgimagetable.bgimage.begin(); bgimage_iterator != p3t_theme.bgimagetable.bgimage.end(); ++bgimage_iterator)
  {
    if (bgimage_iterator->hd.file_data || bgimage_iterator->sd.file_data)
    {
      p3tElement& bgimage = (p3tElement&)buildElement(std::string("bgimage"), bgimagetable.offset, sibling_offset);
      if (previous_element)
      {
        previous_element->header.next_sibling_offset = bgimage.offset;
      }
      else
        bgimagetable.header.first_child_offset = bgimage.offset;
      bgimagetable.header.last_child_offset = bgimage.offset;
      
      if (bgimage_iterator->hd.file_data)
        buildFileAttribute(bgimage, std::string("hd"), bgimage_iterator->hd.file_data, bgimage_iterator->hd.file_size);
      if (bgimage_iterator->sd.file_data)
        buildFileAttribute(bgimage, std::string("sd"), bgimage_iterator->sd.file_data, bgimage_iterator->sd.file_size);
      sibling_offset = bgimage.offset;
      
      previous_element = &bgimage;
    }
  }
  
  p3tElement& font = (p3tElement&)buildElement(std::string("font"), theme.offset, bgimagetable.offset);
  bgimagetable.header.next_sibling_offset = font.offset;
  buildIntegerAttribute(font, std::string("selection"), p3t_theme.font.selection);    
  
  p3tElement& color = (p3tElement&)buildElement(std::string("color"), theme.offset, font.offset);
  font.header.next_sibling_offset = color.offset;
  theme.header.last_child_offset = color.offset;
  buildIntegerAttribute(color, std::string("selection"), p3t_theme.color.selection);   
  
  FILE* p3t_file_handle = fopen(p3t_filename, "wb");
  if (p3t_file_handle == 0)
  {
    return P3T_CANNOT_OPEN_FILE;
  }
  
  p3tHeader p3t_header;
  p3t_header.magic = *((unsigned long*)"P3TF");
  p3t_header.version = 272;
  p3t_header.elements.offset = sizeof(p3t_header);
  p3t_header.elements.size = element_offset;
  p3t_header.ids.offset = ((p3t_header.elements.offset + p3t_header.elements.size + 15) / 16) * 16;
  p3t_header.ids.size = ids_offset;
  p3t_header.strings.offset = ((p3t_header.ids.offset + p3t_header.ids.size + 15) / 16) * 16;
  p3t_header.strings.size = string_offset;
  p3t_header.integers.offset = ((p3t_header.strings.offset + p3t_header.strings.size + 15) / 16) * 16;
  p3t_header.integers.size = 0;
  p3t_header.decimals.offset = ((p3t_header.integers.offset + p3t_header.integers.size + 15) / 16) * 16;
  p3t_header.decimals.size = 0;
  p3t_header.files.offset = ((p3t_header.decimals.offset + p3t_header.decimals.size + 15) / 16) * 16;
  p3t_header.files.size = file_offset;
  p3t_header.padding[0] = 0;
  p3t_header.padding[1] = 0;
  
  fwrite(&p3t_header, 1, sizeof(p3t_header), p3t_file_handle);
  
#ifdef DEBUG_OUTPUT_TEXT
  printf ("File position element: %d\n", ftell(p3t_file_handle));
#endif
  
  for (elementMap::iterator element_iterator = elements.begin(); element_iterator != elements.end(); ++element_iterator)
  {
    element_iterator->second.header.no_attributes = element_iterator->second.attributes.size();
    fwrite(&element_iterator->second.header, 1, sizeof(element_iterator->second.header), p3t_file_handle);
    
    for (p3tAttributeList::const_iterator attribute_iterator = element_iterator->second.attributes.begin(); attribute_iterator != element_iterator->second.attributes.end(); ++attribute_iterator)
    {
      fwrite(&attribute_iterator->header, 1, sizeof(attribute_iterator->header), p3t_file_handle);
    }
  }
  
  char padding = 0;
  for (int padding_index = 0; padding_index < p3t_header.ids.offset - (p3t_header.elements.offset + p3t_header.elements.size); ++padding_index)
  {
    fwrite(&padding, 1, 1, p3t_file_handle);
  }
  
#ifdef DEBUG_OUTPUT_TEXT
  printf ("File position id: %d\n", ftell(p3t_file_handle));
#endif
  
  for (idMap::const_iterator id_iterator = ids.begin(); id_iterator != ids.end(); ++id_iterator)
  {
    p3tOffsetType parentOffset = idParents[id_iterator->first];
    fwrite(&parentOffset, 1, sizeof(parentOffset), p3t_file_handle);
    fwrite(id_iterator->second.c_str(), 1, id_iterator->second.size(), p3t_file_handle);
    fwrite(&padding, 1, 1, p3t_file_handle);
  }
  
  for (int padding_index = 0; padding_index < p3t_header.strings.offset - (p3t_header.ids.offset + p3t_header.ids.size); ++padding_index)
  {
    fwrite(&padding, 1, 1, p3t_file_handle);
  }
  
#ifdef DEBUG_OUTPUT_TEXT
  printf ("File position string: %d\n", ftell(p3t_file_handle));
#endif
  
  for (stringMap::const_iterator string_iterator = strings.begin(); string_iterator != strings.end(); ++string_iterator)
  {
    fwrite(string_iterator->second.c_str(), 1, string_iterator->second.size(), p3t_file_handle);
    fwrite(&padding, 1, 1, p3t_file_handle);
  }
  
  for (int padding_index = 0; padding_index < p3t_header.files.offset - (p3t_header.strings.offset + p3t_header.strings.size); ++padding_index)
  {
    fwrite(&padding, 1, 1, p3t_file_handle);
  }
  
#ifdef DEBUG_OUTPUT_TEXT
  printf ("File position file: %d\n", ftell(p3t_file_handle));
#endif
  
  for (elementMap::const_iterator element_iterator = elements.begin(); element_iterator != elements.end(); ++element_iterator)
  {   
    for (p3tAttributeList::const_iterator attribute_iterator = element_iterator->second.attributes.begin(); attribute_iterator != element_iterator->second.attributes.end(); ++attribute_iterator)
    {
      if (attribute_iterator->header.type == 6)
      {
        fwrite(attribute_iterator->file_data, 1, attribute_iterator->file_size, p3t_file_handle);
        
        p3tSizeType no_file_padding_bytes = 
        ((( (attribute_iterator->header.value.location.offset + attribute_iterator->header.value.location.size) + 15) / 16) * 16) - 
        (attribute_iterator->header.value.location.offset + attribute_iterator->header.value.location.size);
        for (int i = 0; i < no_file_padding_bytes; ++i)
        {
          fwrite(&padding, 1, 1, p3t_file_handle);
        }
        
        if (attribute_iterator->file_type == P3T_GIM_FILE)
        {
          delete[](char*)attribute_iterator->file_data;
        }
      }
    }
  }
  
#ifdef DEBUG_OUTPUT_TEXT
  printf ("File position end: %d\n", ftell(p3t_file_handle));
#endif
  
  fclose(p3t_file_handle);
  
  return P3T_OK;
}

void p3tTheme::clear()
{
  infotable.info.comment = "";
  infotable.info.previewsize = 0;
  infotable.info.name = "";
  infotable.info.author = "";
  infotable.info.url = "";
  infotable.info.authoriconsize = 0;
  if (infotable.info.authoricon.file_data)
    delete infotable.info.authoricon.file_data;
  infotable.info.authoricon.file_data = 0;
  infotable.info.authoricon.file_size = 0;
  infotable.info.iconsize = 0;
  infotable.info.version = "";
  infotable.info.mtime = "";
  infotable.info.genre = "";
  if (infotable.info.preview.file_data)
    delete infotable.info.preview.file_data;
  infotable.info.preview.file_data = 0;
  infotable.info.preview.file_size = 0;
  if (infotable.info.icon.file_data)
    delete infotable.info.icon.file_data;
  infotable.info.icon.file_data = 0;
  infotable.info.icon.file_size = 0;  
  
  infotable.localizedinfo.clear();
  
  for (p3tSoundEffectList::iterator se_iterator = setable.se.begin(); se_iterator != setable.se.end(); ++se_iterator)
  {
    if (se_iterator->left.file_data)
      delete se_iterator->left.file_data;
    if (se_iterator->right.file_data)
      delete se_iterator->right.file_data;
  }
  setable.se.clear();
  
  for (p3tIconList::iterator icon_iterator = icontable.icon.begin(); icon_iterator != icontable.icon.end(); ++icon_iterator)
  {
    if (icon_iterator->src.file_data)
      delete icon_iterator->src.file_data;
  }
  icontable.icon.clear();
  
  for (p3tPointerList::iterator pointer_iterator = pointertable.pointer.begin(); pointer_iterator != pointertable.pointer.end(); ++pointer_iterator)
  {
    if (pointer_iterator->src.file_data)
      delete pointer_iterator->src.file_data;
  }
  pointertable.pointer.clear();
  
  if (notification.src.file_data)
  {
    delete notification.src.file_data;
    notification.src.file_data = 0;
    notification.src.file_size = 0;
    notification.size = 0;
  }
  
  for (p3tBGImageList::iterator bgimage_iterator = bgimagetable.bgimage.begin(); bgimage_iterator != bgimagetable.bgimage.end(); ++bgimage_iterator)
  {
    if (bgimage_iterator->hd.file_data)
      delete bgimage_iterator->hd.file_data;
    if (bgimage_iterator->sd.file_data)
      delete bgimage_iterator->sd.file_data;
  }
  bgimagetable.bgimage.clear();
  
  font.selection = 0;
  color.selection = 0;
}

p3tLocalizedInfoList::iterator p3tTheme::findLocalizedInfo(p3tId locale)
{
  for (p3tLocalizedInfoList::iterator iterator = infotable.localizedinfo.begin(); iterator != infotable.localizedinfo.end(); ++iterator)
  {
    if (iterator->locale == locale)
      return iterator;
  }
  
  return 0;  
}

p3tSoundEffectList::iterator p3tTheme::findSoundEffect(p3tId id)
{
  for (p3tSoundEffectList::iterator iterator = setable.se.begin(); iterator != setable.se.end(); ++iterator)
  {
    if (iterator->id == id)
      return iterator;
  }
  
  return 0;  
}

p3tIconList::iterator p3tTheme::findIcon(p3tId id)
{
  for (p3tIconList::iterator iterator = icontable.icon.begin(); iterator != icontable.icon.end(); ++iterator)
  {
    if (iterator->id == id)
      return iterator;
  }
  
  return 0;  
}

p3tPointerList::iterator p3tTheme::findPointerIcon(p3tId id)
{
  for (p3tPointerList::iterator iterator = pointertable.pointer.begin(); iterator != pointertable.pointer.end(); ++iterator)
  {
    if (iterator->id == id)
      return iterator;
  }
  
  return 0;
}
